Unit INI;
{ ---------------------------------------------------------------------------- }
{ Unit for handling windows-like .INI files -- adapted from Borland Delphi 4.0 }
{                                                                              }
{ Notes                                                                        }
{                                                                              }
{   Methods ReadDate, ReadTDateTime, and ReadTime are not implemented since     }
{   they would require extensive parsing.                                      }
{ ---------------------------------------------------------------------------- }

Interface

{$IFDEF WINDOWS}
Uses Objects,WinDos;
{$ELSE}
Uses Objects,Dos;
{$ENDIF}

Type
{$IFNDEF WINDOWS}
  TDateTime = DateTime;
{$ENDIF

  { Support types--needed because we have no TStringList or TStrings }

  PKeyValue = ^TKeyValue;
  TKeyValue = Object(TObject)
    Key   : PChar;
    Value : PChar;
    Constructor Init(Const _Key,_Value: String);
    Destructor  Done; Virtual;
  End;
  PKeyValueCollection = ^TKeyValueCollection;
  TKeyValueCollection = Object(TSortedCollection)
    Function Compare(Key1,Key2: Pointer): Integer; Virtual;
    Function KeyOf(Item: Pointer): Pointer; Virtual;
  End;
  PINISection = ^TINISection;
  TINISection = Object(TObject)
    Name   : PChar;
    Keys   : PKeyValueCollection;
    Constructor Init(Const _Name: String);
    Destructor  Done; Virtual;
  End;
  PINISectionCollection = ^TINISectionCollection;
  TINISectionCollection = Object(TSortedCollection)
    Function Compare(Sec1,Sec2: Pointer): Integer; Virtual;
    Function KeyOf(Item: Pointer): Pointer; Virtual;
  End;

  { TINIFile Object }

  PINIFile = ^TINIFile;
  TINIFile = Object(TObject)
  Private
    FileName : String;
  Public
    Sections : PINISectionCollection;
    Changed  : Boolean;
    Procedure   DeleteKey(Const Section,Ident: String);
    Destructor  Done; Virtual;
    Procedure   EraseSection(Const Section: String);
    Procedure   ForEachKeyInSection(Const Section: String; Action: Pointer);
    Constructor Init(Const _FileName: String);
    Function    ReadBool(Const Section,Ident: String; Default: Boolean): Boolean;
    Function    ReadFloat(Const Section,Ident: String; Default: Double): Double;
    Function    ReadInteger(Const Section,Ident: String; Default: LongInt): LongInt;
    Procedure   ReadSection(Const Section: String; StrColl: PStrCollection);
    Procedure   ReadSections(StrColl: PStrCollection);
    Procedure   ReadSectionValues(Const Section: String; StrColl: PKeyValueCollection);
    Function    ReadString(Const Section,Ident,Default: String): String;
    Function    SectionExists(Const Section: String): Boolean;
    Procedure   UpdateFile;
    Function    ValueExists(Const Section,Ident: String): Boolean;
    Procedure   WriteBool(Const Section,Ident: String; Const Value: Boolean);
    Procedure   WriteDate(Const Section,Ident: String; Const Value: TDateTime);
    Procedure   WriteDateTime(Const Section,Ident: String; Const Value: TDateTime);
    Procedure   WriteFloat(Const Section,Ident: String; Const Value: Double);
    Procedure   WriteInteger(Const Section,Ident: String; Const Value: LongInt);
    Procedure   WriteString(Const Section,Ident,Value: String);
    Procedure   WriteTime(Const Section,Ident: String; Const Value: TDateTime);
  End;

Function StrPCopyNew(Const St: String): PChar;

Implementation

Uses Strings;

Function StrPCopyNew(Const St: String): PChar;
Var P: PChar;
Begin
  GetMem(P,Length(St) + 1);
  StrPCopy(P,St);
  StrPCopyNew := P;
End; { StrPCopyNew }

Function FileExists(FileName: String): Boolean;
Var Fil: File;
Begin
  Assign(Fil,FileName);
  {$I-}
  Reset(Fil);
  {$I+}
  If (IOResult = 0) Then
  Begin
    Close(Fil);
    FileExists := True;
  End
  Else FileExists := False;
End; { FileExists }

Function Trim(St: String): String;
Begin
  While (St <> '') And (St[1]          = ' ') Do St := Copy(St,2,Length(St) - 1);
  While (St <> '') And (St[Length(St)] = ' ') Do St := Copy(St,1,Length(St) - 1);
  Trim := St;
End; { Trim }

Function UpperCase(St: String): String;
Var I: Byte;
Begin
  If St = '' Then
  Begin
    UpperCase := '';
    Exit;
  End;
  For I := 1 To Length(St) Do St[I] := UpCase(St[I]);
  UpperCase := St;
End; { UpperCase }

{ ------------------------------------------------------- }
{ TKeyValue                                               }
{ ------------------------------------------------------- }

Constructor TKeyValue.Init(Const _Key,_Value: String);
Begin
  Inherited Init;
  Key   := StrPCopyNew(_Key);
  Value := StrPCopyNew(_Value);
End; { TKeyValue.Init }

Destructor TKeyValue.Done;
Begin
  StrDispose(Key);
  StrDispose(Value);
  Inherited Done;
End; { TKeyValue.Done }

{ ------------------------------------------------------- }
{ TKeyValueCollection                                     }
{ ------------------------------------------------------- }

Function TKeyValueCollection.KeyOf(Item: Pointer): Pointer;
Begin
  KeyOf := PKeyValue(Item)^.Key;
End; { TKeyValueCollection.KeyOf }

Function TKeyValueCollection.Compare(Key1,Key2: Pointer): Integer;
Var I: Integer;
Begin
  Compare := StrComp(Key1,Key2);
End; { TKeyValueCollection.Compare }

{ ------------------------------------------------------- }
{ TINISection                                             }
{ ------------------------------------------------------- }

Constructor TINISection.Init(Const _Name: String);
Begin
  Inherited Init;
  Name := StrPCopyNew(_Name);
  Keys := New(PKeyValueCollection,Init(10,10));
End; { TINISection.Init }

Destructor TINISection.Done;
Begin
  Dispose(Keys,Done);
  StrDispose(Name);
  Inherited Done;
End; { TINISection.Done }

{ ------------------------------------------------------- }
{ TINISectionCollection                                   }
{ ------------------------------------------------------- }

Function TINISectionCollection.KeyOf(Item: Pointer): Pointer;
Begin
  KeyOf := PINISection(Item)^.Name;
End; { TINISectionCollection.KeyOf }

Function TINISectionCollection.Compare(Sec1,Sec2: Pointer): Integer;
Var I : Integer;
Begin
  Compare := StrComp(Sec1,Sec2);
End; { TINISectionCollection.Compare }

{ ------------------------------------------------------- }
{ TINIFile                                                }
{ ------------------------------------------------------- }

Procedure TINIFile.DeleteKey(Const Section,Ident: String);
Var
  _Section : PChar;
  _Ident   : PChar;
  SecIndex : Integer;
  KeyIndex : Integer;
  Sec      : PINISection;

Begin
  _Section := StrPCopyNew(UpperCase(Section));
  _Ident   := StrPCopyNew(UpperCase(Ident));
  If Sections^.Search(_Section,SecIndex) Then
  Begin
    Sec := PINISection(Sections^.At(SecIndex));
    If Sec^.Keys^.Search(_Ident,KeyIndex) Then
    Begin
      Sec^.Keys^.AtDelete(KeyIndex);
      Changed := True;
    End;
  End;
  StrDispose(_Section);
  StrDispose(_Ident);
End; { TINIFile.DeleteKey }

Destructor TINIFile.Done;
Begin
  UpdateFile;
  Dispose(Sections,Done);
  Inherited Done;
End; { TINIFile.Done }

Procedure TINIFile.EraseSection(Const Section: String);
Var
  _Section : PChar;
  SecIndex : Integer;

Begin
  _Section := StrPCopyNew(UpperCase(Section));
  If Sections^.Search(_Section,SecIndex) Then
  Begin
    Sections^.AtDelete(SecIndex);
    Changed := True;
  End;
  StrDispose(_Section);
End; { TINIFile.EraseSection }

Procedure TINIFile.ForEachKeyInSection(Const Section: String; Action: Pointer);
Var
  _Section : PChar;
  SecIndex : Integer;
  Sec      : PINISection;

Begin
  _Section := StrPCopyNew(UpperCase(Section));
  If Sections^.Search(_Section,SecIndex) Then
  Begin
    Sec := PINISection(Sections^.At(SecIndex));
    Sec^.Keys^.ForEach(Action);
    Changed  := True;
  End;
  StrDispose(_Section);
End; { TINIFile.ForEachKeyInSection }

Constructor TINIFile.Init(Const _FileName: String);
Var
  F   : Text;
  St  : String;
  I   : Integer;
  Sec : PINISection;

Begin
  Inherited Init;
  Sections := New(PINISectionCollection,Init(10,10));
  FileName := _FileName;
  If FileExists(_FileName) Then
  Begin

    { Read the entire .INI file contents into memory }

    Sec := Nil;
    Assign(F,FileName);
    Reset(F);
    While Not Eof(F) Do
    Begin
      ReadLn(F,St);
      St := Trim(St);
      I  := Pos('=',St);
      If (I > 1) And (Sec <> Nil) And (I < Length(St)) Then
      Begin
        Sec^.Keys^.Insert(New(PKeyValue,Init(UpperCase(Trim(Copy(St,1,I - 1))),Trim(Copy(St,I + 1,Length(St) - I)))));
      End
      Else
      Begin
        If (Copy(St,1,1) = '[') And (Copy(St,Length(St),1) = ']') And (Length(St) > 2) Then
        Begin
          Sec := New(PINISection,Init(UpperCase(Trim(Copy(St,2,Length(St) - 2)))));
          Sections^.Insert(Sec);
        End;
      End;
    End; { While }
    Close(F);
  End;
  Changed := False;
End; { TINIFile.Init }

Function TINIFile.ReadBool(Const Section,Ident: String; Default: Boolean): Boolean;
Var St: String;
Begin
  St := ReadString(Section,Ident,'');
       If St = '0' Then ReadBool := False
  Else If St = '1' Then ReadBool := True
  Else ReadBool := Default;
End; { TINIFile.ReadBool }

Function TINIFile.ReadFloat(Const Section,Ident: String; Default: Double): Double;
Var
  St : String;
  I  : Double;
  J  : Integer;

Begin
  St := ReadString(Section,Ident,'');
  Val(St,I,J);
  If J = 0 Then ReadFloat := I Else ReadFloat := Default;
End; { TINIFile.ReadFloat }

Function TINIFile.ReadInteger(Const Section,Ident: String; Default: LongInt): LongInt;
Var
  St : String;
  I  : LongInt;
  J  : Integer;

Begin
  St := ReadString(Section,Ident,'');
  Val(St,I,J);
  If J = 0 Then ReadInteger := I Else ReadInteger := Default;
End; { TINIFile.ReadInteger }

Procedure TINIFile.ReadSection(Const Section: String; StrColl: PStrCollection);
Var
  I        : Integer;
  _Section : PChar;
  SecIndex : Integer;
  Sec      : PINISection;

Begin
  StrColl^.FreeAll;
  _Section := StrPCopyNew(UpperCase(Section));
  If Sections^.Search(_Section,SecIndex) Then
  Begin
    Sec := PINISection(Sections^.At(SecIndex));
    I := 0;
    While I < Sec^.Keys^.Count Do
    Begin
      StrColl^.Insert(StrNew(PKeyValue(Sec^.Keys^.At(I))^.Key));
      Inc(I);
    End; { While }
  End;
  StrDispose(_Section);
End; { TINIFile.ReadSection }

Procedure TINIFile.ReadSections(StrColl: PStrCollection);
Var I: Integer;
Begin
  StrColl^.FreeAll;
  I := 0;
  While I < Sections^.Count Do
  Begin
    StrColl^.Insert(StrNew(PINISection(Sections^.At(I))^.Name));
    Inc(I);
  End; { While }
End; { TINIFile.ReadSections }

Procedure TINIFile.ReadSectionValues(Const Section: String; StrColl: PKeyValueCollection);
Var
  I        : Integer;
  _Section : PChar;
  SecIndex : Integer;
  Sec      : PINISection;
  Key      : PKeyValue;

Begin
  StrColl^.FreeAll;
  _Section := StrPCopyNew(UpperCase(Section));
  If Sections^.Search(_Section,SecIndex) Then
  Begin
    Sec := PINISection(Sections^.At(SecIndex));
    I := 0;
    While I < Sec^.Keys^.Count Do
    Begin
      Key := PKeyValue(Sec^.Keys^.At(I));
      StrColl^.Insert(New(PKeyValue,Init(StrPas(Key^.Key),StrPas(Key^.Value))));
      Inc(I);
    End; { While }
  End;
  StrDispose(_Section);
End; { TINIFile.ReadSectionValues }

Function TINIFile.ReadString(Const Section,Ident,Default: String): String;
Var
  _Section : PChar;
  _Ident   : PChar;
  SecIndex : Integer;
  KeyIndex : Integer;
  Sec      : PINISection;
  KeyValue : PKeyValue;

Begin
  _Section := StrPCopyNew(UpperCase(Section));
  _Ident   := StrPCopyNew(UpperCase(Ident));
  If Sections^.Search(_Section,SecIndex) Then
  Begin
    Sec := PINISection(Sections^.At(SecIndex));
    If Sec^.Keys^.Search(_Ident,KeyIndex) Then
    Begin
      KeyValue   := Sec^.Keys^.At(KeyIndex);
      ReadString := StrPas(KeyValue^.Value);
    End
    Else ReadString := Default;
  End
  Else ReadString := Default;
  StrDispose(_Section);
  StrDispose(_Ident);
End; { TINIFile.ReadString }

Function TINIFile.SectionExists(Const Section: String): Boolean;
Var
  _Section : PChar;
  SecIndex : Integer;

Begin
  _Section      := StrPCopyNew(UpperCase(Section));
  SectionExists := Sections^.Search(_Section,SecIndex);
  StrDispose(_Section);
End; { TINIFile.SectionExists }

Procedure TINIFile.UpdateFile;
Var
  F   : Text;
  I,J : Integer;
  Sec : PINISection;
  Key : PKeyValue;

Begin
  If Changed Then
  Begin
    Assign(F,FileName);
    ReWrite(F);
    I := 0;
    While I < Sections^.Count Do
    Begin
      Sec := PINISection(Sections^.At(I));
      WriteLn(F,'[',StrPas(Sec^.Name),']');
      J := 0;
      While J < Sec^.Keys^.Count Do
      Begin
        Key := PKeyValue(Sec^.Keys^.At(J));
        WriteLn(F,StrPas(Key^.Key),'=',StrPas(Key^.Value));
        Inc(J);
      End; { While }
      If I < Sections^.Count Then WriteLn(F);
      Inc(I);
    End; { While }
    Close(F);
  End;
End; { TINIFile.UpdateFile }

Function TINIFile.ValueExists(Const Section,Ident: String): Boolean;
Var
  _Section : PChar;
  _Ident   : PChar;
  SecIndex : Integer;
  KeyIndex : Integer;
  Sec      : PINISection;

Begin
  _Section := StrPCopyNew(UpperCase(Section));
  _Ident   := StrPCopyNew(UpperCase(Ident));
  If Sections^.Search(_Section,SecIndex) Then
  Begin
    Sec         := PINISection(Sections^.At(SecIndex));
    ValueExists := Sec^.Keys^.Search(_Ident,KeyIndex);
  End
  Else ValueExists := False;
  StrDispose(_Section);
  StrDispose(_Ident);
End; { TINIFile.ValueExists }

Procedure TINIFile.WriteBool(Const Section,Ident: String; Const Value: Boolean);
Var St: String;
Begin
  If Value Then St := '1' Else St := '0';
  WriteString(Section,Ident,St);
End; { TINIFile.WriteBool }

Procedure TINIFile.WriteDate(Const Section,Ident: String; Const Value: TDateTime);
Var St,St1: String;
Begin
  Str(Value.Month,St);
  St := St + '/';
  Str(Value.Day,St1);
  St := St + St1 + '/';
  Str(Value.Year,St1);
  St := St + St1;
  WriteString(Section,Ident,St);
End; { TINIFile.WriteDate }

Procedure TINIFile.WriteDateTime(Const Section,Ident: String; Const Value: TDateTime);
Var St,St1: String;
Begin

  { Date part }

  Str(Value.Month,St);
  St := St + '/';
  Str(Value.Day,St1);
  St := St + St1 + '/';
  Str(Value.Year,St1);
  St := St + St1 + ' ';

  { Time part }

  Str(Value.Hour,St1);
  St := St + St1 + ':';
  Str(Value.Min,St1);
  St := St + St1 + ':';
  Str(Value.Sec,St1);
  St := St + St1;

  WriteString(Section,Ident,St);
End; { TINIFile.WriteDateTime }

Procedure TINIFile.WriteFloat(Const Section,Ident: String; Const Value: Double);
Var St: String;
Begin
  Str(Value,St);
  WriteString(Section,Ident,St);
End; { TINIFile.WriteFloat }

Procedure TINIFile.WriteInteger(Const Section,Ident: String; Const Value: LongInt);
Var St: String;
Begin
  Str(Value,St);
  WriteString(Section,Ident,St);
End; { TINIFile.WriteInteger }

Procedure TINIFile.WriteString(Const Section,Ident,Value: String);
Var
  _Section : PChar;
  _Ident   : PChar;
  SecIndex : Integer;
  KeyIndex : Integer;
  Sec      : PINISection;
  KeyValue : PKeyValue;

Begin
  _Section := StrPCopyNew(UpperCase(Section));
  _Ident   := StrPCopyNew(UpperCase(Ident));
  If Sections^.Search(_Section,SecIndex) Then
  Begin
    Sec := PINISection(Sections^.At(SecIndex));
    If Sec^.Keys^.Search(_Ident,KeyIndex) Then
    Begin
      KeyValue := Sec^.Keys^.At(KeyIndex);
      StrDispose(KeyValue^.Value);
      KeyValue^.Value := StrPCopyNew(Value);
    End
    Else
    Begin
      KeyValue := New(PKeyValue,Init(Ident,Value));
      Sec^.Keys^.Insert(KeyValue);
    End;
  End
  Else
  Begin
    Sec := New(PINISection,Init(UpperCase(Section)));
    Sections^.Insert(Sec);
    KeyValue := New(PKeyValue,Init(UpperCase(Ident),Value));
    Sec^.Keys^.Insert(KeyValue);
  End;
  Changed := True;
  StrDispose(_Section);
  StrDispose(_Ident);
End; { TINIFile.WriteString }

Procedure TINIFile.WriteTime(Const Section,Ident: String; Const Value: TDateTime);
Var St,St1: String;
Begin
  Str(Value.Hour,St);
  St := St + ':';
  Str(Value.Min,St1);
  St := St + St1 + ':';
  Str(Value.Sec,St1);
  St := St + St1;
  WriteString(Section,Ident,St);
End; { TINIFile.WriteTime }

End.
